/*************************************************************************/
/*                                                                       */
/*  libservices - /etc/services handling library.                        */
/*                                                                       */
/*  Written by Gustavo Niemeyer <gustavo@niemeyer.net>                   */
/*                                                                       */
/*  Copyright (C) 2001 Conectiva, Inc.                                   */
/*                                                                       */
/*  This library is free software; you can redistribute it and/or        */
/*  modify it under the terms of the GNU Library General Public          */
/*  License as published by the Free Software Foundation; either         */
/*  version 2 of the License, or (at your option) any later version.     */
/*                                                                       */
/*  This library is distributed in the hope that it will be useful,      */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    */
/*  Library General Public License for more details.                     */
/*                                                                       */
/*  You should have received a copy of the GNU Library General Public    */
/*  License along with this library; if not, write to the                */
/*  Free Software Foundation, Inc., 59 Temple Place - Suite 330,         */
/*  Boston, MA  02111-1307, USA.                                         */
/*                                                                       */
/*************************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include "libservices.h"

enum {
  ACT_SHOW,
  ACT_ADD,
  ACT_REMOVE,
  ACT_REPLACE,
  ACT_VERSION,
  ACT_HELP
};

struct options {
  char action;
  char *name;
  char *port;
  char *proto;
  struct serv_alias *aliases;
  char comment;
  char *comments;
  char verbose;
  char force;
};

static int parse_options(int argc, char *argv[], struct options *opt)
{
  struct serv_alias *alias = NULL, *new_alias = NULL;
  int c;
  memset(opt, 0, sizeof(*opt));
  opt->action = ACT_SHOW;
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"show", 0, 0, ACT_SHOW},
      {"add", 0, 0, ACT_ADD},
      {"remove", 0, 0, ACT_REMOVE},
      {"replace", 0, 0, ACT_REPLACE},
      {"name", 1, 0, 'n'},
      {"port", 1, 0, 'p'},
      {"proto", 1, 0, 'o'},
      {"alias", 1, 0, 'a'},
      {"comments", 2, 0, 'c'},
      {"verbose", 0, 0, 'v'},
      {"force", 0, 0, 'f'},
      {"version", 0, 0, ACT_VERSION},
      {"help", 0, 0, ACT_HELP},
      {0, 0, 0, 0}
    };
    c = getopt_long (argc, argv, "n:p:o:a:c::", long_options, &option_index);
    if (c == -1) break;
    if (c >= ACT_SHOW && c <= ACT_HELP)
      opt->action = c;
    else switch (c) {
      case 'n':
        opt->name = optarg;
        break;
      
      case 'p':
        opt->port = optarg;
        break;
      
      case 'o':
        opt->proto = optarg;
        break;
      
      case 'a':
        new_alias = (struct serv_alias *) malloc(sizeof(struct serv_alias));
        if (!new_alias) {
          fprintf(stderr,"error: can't allocate alias structure\n");
          return 1;
        }
        new_alias->alias = strdup(optarg);
        if (!alias) {
          opt->aliases = alias = new_alias;
        } else {
          alias->next = new_alias;
          new_alias->prev = alias;
          alias = new_alias;
        }
        break;
      
      case 'c':
        opt->comment = 1;
        opt->comments = optarg;
        break;
      
      case 'v':
        opt->verbose = 1;
        break;
      
      case 'f':
        opt->force = 1;
        break;
      
      case '?':
        return 1;
        break;
    }
  }
  if (optind < argc) {
    fprintf(stderr, "error: invalid parameters:");
    while (optind < argc)
      fprintf(stderr, " %s", argv[optind++]);
    fprintf(stderr, "\n");
    return 1;
  }
  return 0;
}

static void print_version()
{
  printf(
    "services (libservices) %s\n"
    "Written by Gustavo Niemeyer.\n"
    "Copyright (C) 2001 Conectiva, Inc.\n\n"
    , VERSION);
}

static void print_usage(char *argv[])
{
  printf(
    "Usage: %s [ACTION] [OPTION]...\n"
    "Append, remove, and change services settings.\n"
    "\n"
    "Actions:\n"
    "      --show                  show service(s) (default action)\n"
    "      --add                   add service (won't replace)\n"
    "      --remove                remove service(s)\n"
    "      --replace               add service and replace if necessary\n"
    "\n"
    "Options:\n"
    "  -n, --name=STRING           service name\n"
    "  -p, --port=NUMBER           service port number\n"
    "  -o, --proto=STRING          service protocol\n"
    "  -a, --alias=STRING          service alias\n"
    "      --comments[=STRING]     display/add/replace comments\n"
    "      --verbose               gives more information\n"
    "      --force                 force a dangerous action\n"
    "      --help                  display this help and exit\n"
    "      --version               output version information and exit\n"
    "\n"
    , argv[0]
  );
}

static int show_services(const struct options *opt)
{
  struct serv *serv_list = serv_load(NULL);
  struct serv *serv = serv_list;
  struct serv_alias *alias;
  int ret = 1;
  for (; serv; serv = serv->next) {
    if ((!opt->name || strcmp(opt->name, serv->name)==0)
        && (!opt->aliases || serv_match_any_alias(serv->aliases, opt->aliases))
        && (!opt->port || strcmp(opt->port, serv->port)==0)
        && (!opt->proto || strcmp(opt->proto, serv->proto) == 0)) {
      printf("%s%s%s\t%s", serv->name, (strlen(serv->name)<8)?"\t\t":"\t",
             serv->port, serv->proto);
      alias = serv->aliases;
      if (alias) {
        printf("\t%s", alias->alias);
        alias = alias->next;
        for (; alias; alias = alias->next)
          printf(" %s", alias->alias);
      }
      if (opt->comment && serv->comments)
        printf("\t# %s", serv->comments);
      printf("\n");
      ret = 0;
    }
  }
  serv_destroy(serv_list);
  return ret;
}

static int remove_services(const struct options *opt)
{
  struct serv *serv_list = serv_load(NULL);
  struct serv *serv = serv_list;
  struct serv_alias *alias;
  int ret = 1;
  if (!opt->force && (!opt->name && !opt->port && !opt->aliases)) {
    fprintf(stderr, "error: use '--force' if you really want to do this\n");
    return 1;
  }
  for (; serv; serv = serv->next) {
    if ((!opt->name || strcmp(opt->name, serv->name)==0)
        && (!opt->aliases || serv_match_any_alias(serv->aliases, opt->aliases))
        && (!opt->port || strcmp(opt->port, serv->port)==0)
        && (!opt->proto || strcmp(opt->proto, serv->proto) == 0)) {
      if (opt->verbose) {
        printf("removing %s%s%s\t%s", serv->name, (strlen(serv->name)<7)?"\t\t":"\t",
               serv->port, serv->proto);
        alias = serv->aliases;
        if (alias) {
          printf("\t%s", alias->alias);
          alias = alias->next;
          for (; alias; alias = alias->next)
            printf(" %s", alias->alias);
        }
        if (opt->comment && serv->comments)
          printf("\t# %s", serv->comments);
        printf("\n");
      }
      serv->flags = SERV_REMOVED;
      ret = 0;
    }
  }
  if (ret == 0) {
    ret = serv_write(NULL, serv_list);
    if (ret != 0)
      fprintf(stderr, "error: can't write changes\n");
  } else {
    fprintf(stderr, "warning: no services removed\n");
  }
  serv_destroy(serv_list);
  return ret;
}

static int add_service(const struct options *opt, int replace)
{
  int ret = 1;
  if (!opt->name || !opt->port || !opt->proto) {
    fprintf(stderr, "error: you must provide the name, port, and protocol at least\n");
  } else {
    struct serv *serv_list = serv_load(NULL);
    struct serv *serv = serv_list;
    struct serv_alias *alias;
    char add = 1;
    for (; serv; serv = serv->next) {
      if ((strcmp(opt->port, serv->port)==0)
          && (strcmp(opt->proto, serv->proto) == 0)
          && (!opt->aliases || serv_match_any_alias(serv->aliases, opt->aliases))) {
        add = 0;
        if (!replace) {
          fprintf(stderr, "error: resources already used by %s %s/%s\n",
                  serv->name, serv->port, serv->proto);
        } else {
          serv_destroy_aliases(serv->aliases);
          serv->aliases = opt->aliases;
          serv->comments = opt->comment?strdup(opt->comments?opt->comments:""):NULL;
          ret = serv_write(NULL, serv_list);
          if (ret != 0)
            fprintf(stderr, "error: can't write changes\n");
        }
      }
    }
    if (add) {
      struct serv *serv_new = serv_create_new();
      if (!serv_new) {
        fprintf(stderr, "error: can't create new service\n");
      } else {
        serv_new->name = strdup(opt->name);
        serv_new->port = strdup(opt->port);
        serv_new->proto = strdup(opt->proto);
        serv_new->aliases = opt->aliases;
        if (opt->comment)
          serv_new->comments = strdup(opt->comments?opt->comments:"");
        serv_add(serv_list, serv_new);
        ret = serv_write(NULL, serv_list);
        if (ret != 0)
          fprintf(stderr, "error: can't write changes\n");
      }
    }
    serv_destroy(serv_list);
  }
  return ret;
}

int main(int argc, char *argv[])
{
  struct options opt;
  int ret = 1;
  if (parse_options(argc, argv, &opt) == 0) {
    switch (opt.action) {
      case ACT_SHOW:
        ret = show_services(&opt);
        break;
      case ACT_REMOVE:
        ret = remove_services(&opt);
        break;
      case ACT_ADD:
        ret = add_service(&opt,0);
        break;
      case ACT_REPLACE:
        ret = add_service(&opt,1);
        break;
      case ACT_VERSION:
        print_version();
        ret = 0;
        break;
      case ACT_HELP:
        print_usage(argv);
        ret = 0;
        break;
    }
  } else {
    print_usage(argv);
  }
  return ret;
}

/* vim:sw=2:ts=2:et
*/
